/*
 * File name: pctspn_som.cc
 * Date:      2013/10/13 09:02
 * Author:    Jan Faigl
 */

#include <cmath>
#include <limits>
#include <iomanip>

#include <boost/foreach.hpp>

#include "random.h"
#include "timerN.h"
#include "perf_timer.h"
#include "logging.h"
#include "stringconversions.h"
#include "nodeloader.h"
#include "text_result_log.h"
#include "imr_file_utils.h"
#include "imr_assert.h"

#include "shapes.h"
#include "colormap.h"
#include "canvasview_coords.h"

#include "simple_intersection.h"

#include "pctspn_som.h"

using namespace som;
using namespace imr;
using namespace imr::gui;

#define foreach BOOST_FOREACH

typedef imr::geom::CIntersection<Coords> Intersection;

static const double NEURON_COORDS_IDENTITY2= 1e-5*1e-5;

typedef std::vector<int> IntVector;

/// ----------------------------------------------------------------------------
void createPermutation(int number, IntVector & permutation) {
   permutation.clear();
   for (int i = 0; i < number; i++) { 
      permutation.push_back(i); 
   }
}

/// ----------------------------------------------------------------------------
void permute(IntVector & permutation) {
   int k, tmp;
   imr::CRandom::randomize();
   for (int i = permutation.size(); i > 0; --i) {
      k = imr::CRandom::random()%i;
      tmp = permutation[i-1];
      permutation[i-1] = permutation[k];
      permutation[k] = tmp;
   }
}

/// ----------------------------------------------------------------------------
/// Class CPCTSPNSom
/// ----------------------------------------------------------------------------

/// - static method ------------------------------------------------------------
imr::CConfig& CPCTSPNSom::getConfig(imr::CConfig& config) {
   // basic config
   config.add<std::string>("output", "output directory to store particular results and outputs", "./");
   config.add<std::string>("results", "result log file, it will be placed in output directory", "results.log"); //TODO .db for sqlite, .txt text logs
   config.add<std::string>("info", "information file, it will be placed in particular experiment directory", "info.txt");
   config.add<std::string>("settings", "store configurations in boost::program_options config file format ", "settings.txt");
   config.add<std::string>("result-path", "file name for the final found path (ring) as sequence of points", "path.txt");
   config.add<std::string>("result-canvas-output", "file name for final canvas output (eps,png,pdf,svg) are supported");
   config.add<std::string>("result-canvas-suffixes", "comman separated list of exptensis e.g. png,pdf.  if specified  several result images are created, with particular suffix to the resultCanvasOutput");
   config.add<std::string>("name", "name used in result log as user identification if not set a default values (cities) is used");
   config.add<int>("iteration", "set particular interation, otherwise all interations (batch) are performed", -1); 
   config.add<int>("batch", "number of iterations from 0 to batch (not included) ", -1); 
   config.add<bool>("continue", "in batch mode, partial results are loaded and checked, only missing iterations are computed ", false); 
   config.add<bool>("save-results", "disable/enable save results,configs and so on", true); 
   config.add<bool>("save-info", "disable/enable save info", true); 
   config.add<bool>("save-settings", "disable/enable save settings", true); 
   config.add<bool>("save-visual", "disable/enable save visualization results, canvas be", true); 
   config.add<bool>("verbose-result-log", "disable/enable printing results log into logger", false);
   // end basic config
   config.add<double>("learning-rate", "neuron adaptation parametr in activation function mi*exp(d)", 0.6); //0.6
   config.add<double>("number-neurons-multiplication", "multiplication parametr to get number of neurons as multiplication of number of cities", 2.5);
   config.add<double>("gain-decreasing-rate", "Decrasing rate, higher value means faster convergence", 1e-4);
   config.add<double>("neighborhood-factor", "factor to compute number of neighborhood neurons as division of number of neurons, neighborhood are considered on both sides (left, right) of winning neuron so d*2 neurons are moved ", 5);
   config.add<double>("penalty-factor", "Factor to adjust penalty value read from the file penalty = penalty / factor", 1.0);
   config.add<int>("termination-max-steps", "Stop adaptation if number of steps is large than given", 180);
   config.add<double>("termination-minimal-learning-rate", "Stop adaptation if the current learning rate is below this threshold", 0.4);
   config.add<double>("termination-error", "Stop adaptation when the current error is less than this value", 0.001);

   config.add<double>("communication-radius", "Radius within other sensors can be read, disabled if <= 0", 0);
   config.add<bool>("best-path", "Enable/disable considering best path found during the evaluation", false);


   config.add<std::string>("pic-dir", "relative directory in result directory to store pictures from each iteration");
   config.add<std::string>("pic-ext", "extension of pic, eps, png, pdf, svg (supported by particular gui renderer rendered", "png");
   config.add<bool>("save-pic", "enable/disable saving pictures (after each refine)");

   config.add<bool>("draw-stations", "Enable/Disable drawing stations", true);
   config.add<bool>("draw-ring", "Enable/Disable drawing ring in the final shoot", true);
   config.add<bool>("draw-path", "Enable/Disable drawing ring in the final shoot", true);
   config.add<bool>("draw-alternate-goals", "Enable/Disable drawing alternate goals in drawing ring", false);

   config.add<std::string>("draw-shape-stations", "Shape of the station", Shape::CITY());
   config.add<std::string>("draw-shape-neuron", "Shape of the neurons", Shape::NEURON());
   config.add<std::string>("draw-shape-winner", "Shape of the winners", Shape::NEURON());
   config.add<std::string>("draw-shape-depot", "Shape of the depot", Shape::DEPOT());
   config.add<std::string>("draw-shape-path", "Shape of the path", Shape::RED_LINE());
   config.add<std::string>("draw-shape-ring", "Shape of the ring", Shape::GREEN_BOLD_LINE());
   config.add<std::string>("draw-shape-communication-radius", "Shape of the communication radius", Shape::RED_LINE());
   config.add<std::string>("draw-shape-communication-radius-coverage", "Shape of the communication radius for highlight coverage", Shape::POLYGON_FILL());
   config.add<std::string>("draw-shape-tour-represented-by-ring", Shape::BLACK_BOLD_LINE());
   config.add<std::string>("draw-shape-alternate-goal", Shape::MAP_VERTEX());

   config.add<bool>("draw-neurons", "enable/disable drawing neurons", false);
   config.add<bool>("draw-winners", "enable/disable drawing winner using a different shape", false);
   config.add<bool>("draw-ring-iter", "enable/disable drawing ring at each iteration", false);
   config.add<bool>("draw-path-vertices", "enable/disable drawing path vertices(nodes)", true);
   config.add<bool>("draw-coverage", "enable/disable drawing coverage", false);
   config.add<bool>("draw-coverage-last", "enable/disable drawing coverage at the end of the adaptation", false);
   config.add<bool>("draw-covered", "enable/disable enhancing covered stations", false);
   config.add<bool>("draw-stations-penalty", "enable/disable drawing stations in different color using penalty", false);
   config.add<std::string>("draw-stations-penalty-palette", "File name with colors for the penalty palette", "");
   config.add<bool>("draw-communication-radius", "enable/disable drawing radius of the selected targets", false);
   config.add<bool>("draw-winner-communication-radius", "enable/disable drawing radius of the winner", false);
   config.add<bool>("draw-tour-represented-by-ring", "enable/disable drawing tour represented by ring", false);

   config.add<std::string>("input-path", "If this input file is given, a solution is computed from it considering the given radius", "");
   return config;
}

/// - constructor --------------------------------------------------------------
CPCTSPNSom::CPCTSPNSom(imr::CConfig& config, const std::string& problemFile, bool loadPenalty) :
   Base(config), 
   GAIN_DECREASING_RATE(config.get<double>("gain-decreasing-rate")),
   NEIGHBORHOOD_FACTOR(config.get<double>("neighborhood-factor")),
   COMM_RADIUS(config.get<double>("communication-radius")),
   COMM_RADIUS2(COMM_RADIUS * COMM_RADIUS),
   COMM_RADIUS_SMALLER(COMM_RADIUS - 0.1),
   SAVE_RESULTS(config.get<bool>("save-results")),
   SAVE_SETTINGS(config.get<bool>("save-settings")),
   SAVE_INFO(config.get<bool>("save-info")),
   PROCESS_PATH(!config.get<std::string>("input-path").empty())
{
   savePicEnabled = config.get<bool>("save-pic");
   nbrNeurons = 0;
   neurons = 0;

   drawNeurons = config.get<bool>("draw-neurons");
   drawAlternateGoals = config.get<bool>("draw-alternate-goals");
   drawWinners = config.get<bool>("draw-winners");
   drawCommRadius = config.get<bool>("draw-communication-radius");
   drawWinnerCommRadius = config.get<bool>("draw-winner-communication-radius");
   shapeRing.setShape(config.get<std::string>("draw-shape-ring"));
   shapeNeuron.setShape(config.get<std::string>("draw-shape-neuron"));
   shapeWinner.setShape(config.get<std::string>("draw-shape-winner"));
   shapeCommRadius.setShape(config.get<std::string>("draw-shape-communication-radius"));
   shapeCommRadiusCoverage.setShape(config.get<std::string>("draw-shape-communication-radius-coverage"));
   drawRingIter = config.get<bool>("draw-ring-iter");
   drawRingEnabled = config.get<bool>("draw-ring");
   drawCoverageEnable = config.get<bool>("draw-coverage");
   drawCoveredEnable = config.get<bool>("draw-covered");
   drawCoverageLastEnable= config.get<bool>("draw-coverage-last");
   drawTourRepresentedByRing = config.get<bool>("draw-tour-represented-by-ring");
   shapeAlternateGoal.setShape(config.get<std::string>("draw-shape-alternate-goal"));
   shapeTourRepresentedByRing.setShape(config.get<std::string>("draw-shape-tour-represented-by-ring"));

   if (!loadPenalty) {
      load_stations_coords(problemFile);
   } else {
      load_stations_penalty(problemFile);
   }
   ASSERT_ARGUMENT(!stations.empty(), "Empty stations loaded");
   coveredStations.resize(stations.size());
}

/// - destructor ---------------------------------------------------------------
CPCTSPNSom::~CPCTSPNSom() {
   deallocate_neurons();
   releaseStations();
}

/// ----------------------------------------------------------------------------
std::string CPCTSPNSom::getRevision(void) {
   return "$Id$"; 
}

/// ----------------------------------------------------------------------------
void CPCTSPNSom::setTimers(const imr::CTimerN& load, const imr::CTimerN& init) {
   loadTimer = load;
   initTimer = init;
}

/// ----------------------------------------------------------------------------
std::string CPCTSPNSom::getVersion(void) {
   return "CPCTSPNSom 0.8";
}

/// ----------------------------------------------------------------------------
void CPCTSPNSom::solve(void) {
   if (PROCESS_PATH) {
      load();
      initialize();
      processPath();
      release();
   } else {
      imr::CRandom::randomize();
      Base::solve();
   }
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::load(void) {
   // nothing to load, structures are passed to the constructor
   n = stations.size();
   if (canvas) {
      *canvas << canvas::AREA;
      foreach(SStation* station, stations) { *canvas << station->coords; }
      *canvas << canvas::END;
      if (config.get<bool>("draw-stations")) {
	 std::string pallete = config.get<std::string>("draw-stations-penalty-palette");
	 if (config.get<bool>("draw-stations-penalty") and imr::isFile(pallete)) {
	    imr::gui::CColorMap map;
	    map.load(pallete);
	    double maxPenalty = 0;
	    double minPenalty = std::numeric_limits<double>::max();
	    foreach(SStation* s, stations) { 
	       if (maxPenalty < s->penalty) { maxPenalty = s->penalty; }
	       if (minPenalty > s->penalty) { minPenalty = s->penalty; }
	    }
	    if (minPenalty == maxPenalty) { minPenalty = 0.99 * maxPenalty; }
	    map.setRange(minPenalty, maxPenalty);
	    *canvas << "stations" << CShape(config.get<std::string>("draw-shape-stations")) << canvas::POINT;
	    foreach(SStation* g, stations) {
	       *canvas << imr::gui::canvas::FILL_COLOR << map.getColor(g->penalty) << g->coords;
	    }
	 } else {
	    *canvas << "stations" << CShape(config.get<std::string>("draw-shape-stations")) << canvas::POINT;
	    foreach(const SStation* station, stations) { *canvas << station->coords << canvas::END; }
	 }
      }
      canvas->redraw();
   } //end canvas
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::initialize(void) {
   createPermutation(stations.size(), permutation); // create permutation
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::iterate(int iter) {
   allocate_neurons();

   { //initial settings of the learning parameters
      G = 10;
      mi = config.get<double>("learning-rate");
      d = nbrNeurons / NEIGHBORHOOD_FACTOR;
   }

   //thresholds for the termination conditions
   const double MAX_ERROR = config.get<double>("termination-error");
   const double MIN_LEARNING_RATE = config.get<double>("termination-minimal-learning-rate");
   const int MAX_STEPS = config.get<int>("termination-max-steps");

   if (canvas) {
      *canvas << canvas::CLEAR << "path" << "path";
   }

   double error = 2 * MAX_ERROR;
   int step = 0; 

   //this initialization must be done in each iteration
   initialize_neurons(); // place neurons around first station
   foreach(SStation* station, stations) { station->stepWinnerSelected = -1; }

   bool end = false;

   int bestPathStep = -1;
   CoordsVector path;
   CoordsVector bestPath;
   double path_length;
   double path_penalty;
   double bestPathLength = std::numeric_limits<double>::max();
   double bestPathPenalty = std::numeric_limits<double>::max();
   double bestPathCost = std::numeric_limits<double>::max();
   double bestSolutionError;
   int stationsCovered = 0;
   while( !((error < MAX_ERROR) ) && (step < MAX_STEPS) && !end) { //perform adaptation step
      error = refine(step, error);
      get_ring_path(step, path);
      path_length = get_path_length(path);
      path_penalty = get_path_penalty(path, &stationsCovered);
      const double c = path_length + path_penalty;
      if (!path.empty() and bestPathCost > c) {
	 bestPathStep = step;
	 bestPathCost = c;
	 bestPathLength = path_length;
	 bestPathPenalty = path_penalty;
	 bestPath = path;
	 bestSolutionError = error;
      }
      DEBUG("refine step: " << step << " finished error:" << error << " len: " << path_length << " cost: " << c << " bestLen: " << bestPathLength << " bestCost: " << bestPathCost);

      { //update learning parameters 
	 G = G * (1 - GAIN_DECREASING_RATE*step);
	 d = nbrNeurons / NEIGHBORHOOD_FACTOR;
	 mi = 1 / sqrt(sqrt(step + 1 ));
	 if (mi < MIN_LEARNING_RATE) { end = true; } 
      }
      step++;
   } //end step loop
   tSolve.stop(); 
   DEBUG("Best path with the cost: " << bestPathCost << " found in: " << bestPathStep << " steps");

   double pathLength;
   double pathPenalty = 0.0;
   if (config.get<bool>("best-path")) {
      pathLength = bestPathLength;
      pathPenalty = bestPathPenalty;
      path = bestPath;
   } else {
      get_ring_path(step - 1, path); //default log results (iter, length, steps)
      pathLength = get_path_length(path);
      pathPenalty = get_path_penalty(path, &stationsCovered); 
   }
   finalPath = path;
   fillResultRecord(iter, pathLength, pathPenalty, step); //default log results (iter, length, steps)
   double maxPenalty = 0.0;
   resultLog 
      << bestSolutionError
      << nbrNeurons
      << stations.size()
      << stationsCovered
      << maxPenalty
      << config.get<double>("penalty-factor")
      << imr::result::endrec;
   DEBUG("number neurons: " << nbrNeurons << " stations: " << stations.size() << " path: " << finalPath.size() << " len: " << pathLength << " penalty: " << pathPenalty << " cost: " << (pathLength + pathPenalty));

   if (canvas and drawCoverageLastEnable) {
      CoordsVector path;
      std::vector<bool> covered;
      get_path_covered_stations(finalPath, covered);
      *canvas << canvas::ARC << shapeCommRadiusCoverage;
      foreach(const Coords& ct, finalPath) {
	 *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
      }
      *canvas 
	 << canvas::CLEAR << "stations" << "stations"
	 << shapeCommRadiusCoverage << canvas::ARC;
      foreach(const Coords& ct, finalPath) {
	 *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
      }
      *canvas 
	 << CShape(config.get<std::string>("draw-shape-stations")) 
	 << canvas::POINT; 
      for(int gi = 0; gi < covered.size(); ++gi) {
	 *canvas << (covered[gi] ? FillColor("red") : FillColor("gray")) << stations[gi]->coords;
      }
      if (savePicEnabled) { savePic(step, true); }
   } //end drawCoverage
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::save(void) {
   std::string dir;
   updateResultRecordTimes(); //update timers as load and initilization is outside class
   DEBUG("LOAD_TIME_CPU: " << tLoad.cpuTime());
   DEBUG("INIT_TIME_CPU: " << tInit.cpuTime());
   DEBUG("SAVE_TIME_CPU: " << tSave.cpuTime());
   DEBUG("SOLVE_TIME_CPU: " << tSolve.cpuTime());
   if (SAVE_SETTINGS) {
      saveSettings(getOutputIterPath(config.get<std::string>("settings"), dir));
   }
   if (SAVE_INFO) {
      saveInfo(getOutputIterPath(config.get<std::string>("info"), dir));
   }
   if (SAVE_RESULTS) {
      std::string file = getOutputIterPath(config.get<std::string>("result-path"), dir);
      assert_io(createDirectory(dir), "Can not create file in path'" + file + "'");
      std::ofstream ofs(file.c_str());
      assert_io(not ofs.bad(), "Cannot create path '" + file + "'");
      ofs << std::setprecision(14);
      foreach(const Coords& pt, finalPath) {
	 ofs << pt.x << " " << pt.y << std::endl;
      }
      assert_io(not ofs.bad(), "Error occur during path saving");
      ofs.close();
   }
   if (canvas) { // map must be set 
      *canvas << canvas::CLEAR << "ring";
      if (config.get<bool>("draw-path")) {
	 drawPath();
      } else if (drawRingEnabled) {
	 drawRing(-1);
      }
      saveCanvas(); 
   }
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::release(void) {
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::visualize(void) {
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::defineResultLog(void) {
   static bool resultLogInitialized = false;
   if (!resultLogInitialized) {
      resultLog << result::newcol << "NAME";
      resultLog << result::newcol << "METHOD";
      resultLog << result::newcol << "ITER";
      resultLog << result::newcol << "RTIME";
      resultLog << result::newcol << "CTIME";
      resultLog << result::newcol << "UTIME";
      resultLog << result::newcol << "LENGTH";
      resultLog << CResultLog::AddColumn("PENALTY", "x");
      resultLog << CResultLog::AddColumn("COST", "x");
      resultLog << result::newcol << "STEPS";
      resultLog << CResultLog::AddColumn("ERROR", "x");
      resultLog << CResultLog::AddColumn("NBR_NEURONS", "x");
      resultLog << CResultLog::AddColumn("NBR_STATIONS", "x");
      resultLog << CResultLog::AddColumn("NBR_STATIONS_VISITED", "x");
      resultLog << CResultLog::AddColumn("MAX_PENALTY", "x");
      resultLog << CResultLog::AddColumn("PENALTY_FACTOR", "x");
      resultLogInitialized = true;
   }
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::fillResultRecord(int iter, double length, double penalty, int steps) {
   resultLog << result::newrec << name << getMethod() << iter;
   long t[3] = {0, 0, 0};
   tLoad.addTime(t);
   tInit.addTime(t);
   tSolve.addTime(t);
   tSave.addTime(t);
   resultLog << t[0] << t[1] << t[2] << length << penalty << (length + penalty) << steps;
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::allocate_neurons(void) { // allocate neurons
   const double neuronMul = config.get<double>("number-neurons-multiplication");
   const int m = (int)round(stations.size() * neuronMul);
   deallocate_neurons();
   neurons = new SNeuron();
   neurons->next = neurons; neurons->prev = neurons;
   nbrNeurons = 1;
   for (int i = 1; i < m; i++) {
      insertNeuron(neurons, new SNeuron());
   }
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::deallocate_neurons(void)  {// release neurons
   while(nbrNeurons > 0) {
      removeNeuron(neurons->next);
   }
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::initialize_neurons(void) {
   static const double INITIAL_NEURON_WEIGHTS_RADIUS = 0.5;
   if (!stations.empty()) { // place initial neurons weights around the first stations
      SNeuron* n = neurons;
      const double pom = 2 * M_PI / nbrNeurons;
      const Coords& depot = stations.front()->coords;
      for(int i = 0; i < nbrNeurons; ++i) {
	 n->setPoint(Coords(
		  depot.x + INITIAL_NEURON_WEIGHTS_RADIUS * cos(i * pom),
		  depot.y + INITIAL_NEURON_WEIGHTS_RADIUS * sin(i * pom)
		  ));
	 n = n->next;
      } //end initial number of neurons
   }
}

/// - private method -----------------------------------------------------------
CPCTSPNSom::SNeuron* CPCTSPNSom::insertNeuron(SNeuron* cur, SNeuron* neuron) {
   // add neuron after the cur neuron
   neuron->prev = cur;
   neuron->next = cur->next;
   cur->next = neuron;
   neuron->next->prev = neuron;
   nbrNeurons++;
   return neuron;
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::removeNeuron(SNeuron* neuron) {
   neuron->prev->next = neuron->next;
   neuron->next->prev = neuron->prev;
   if (neurons == neuron) {
      neurons = neuron->next;
      if (nbrNeurons == 1) {
	 neurons = 0;
      }
   }
   delete neuron;
   nbrNeurons--;
}

/// - protected method ---------------------------------------------------------
void CPCTSPNSom::after_init(void) {
   tLoad.append(loadTimer);
   tInit.append(initTimer);
}

/// - protected method ---------------------------------------------------------
double CPCTSPNSom::refine(int step, double errorMax) {
   double errorToGoal = errorMax;
   double error = 0.0;
   permute(permutation);

   for(int i = 0; i < coveredStations.size(); ++i) {
      coveredStations[i] = false;
   }

   //select winner neuron to each city
   for (IntVector::iterator i = permutation.begin(); i != permutation.end(); i++) {
      SStation* station = stations[*i];
      if (coveredStations[*i]) { continue; } //skip already covered station
      if (SNeuron* neuron = selectWinner(step, station, errorToGoal)) { // winner neuron has been selected
	 station->stepWinnerSelected = step;
	 adapt(neuron, step);
	 for(int si = 0; si < stations.size(); ++si) { //consider coverage from the alternate station
	    if (neuron->alternateGoal.squared_distance(stations[si]->coords) <= COMM_RADIUS2) {
	       coveredStations[si] = true;
	    }
	 } //end stations loop
	 if (error < errorToGoal) {
	    error = errorToGoal; //update error
	 }
      } // end adaptation
   } //end permutation of all stations

   ring_regenerate(step);
   if (canvas and drawRingIter) {
      if (drawCoverageEnable) {
      }
      drawRing(step);
      if (savePicEnabled) { savePic(step, true); }
   }
   return error; // return largest error to city
}

/// - private method -----------------------------------------------------------
CPCTSPNSom::SNeuron* CPCTSPNSom::selectWinner(int step, SStation* station, double& errorToGoal) {
   SNeuron* winner = 0;
   int winnerIDX = -1;
   SNeuron* cur = neurons;
   Coords newPoint;
   SNeuron* preWinner = 0;
   Coords x;
   char s;
   bool hasWinner = false;
   double bestCost = std::numeric_limits<double>::max();
   double error = std::numeric_limits<double>::max();
   Coords bestAlternateGoal;
   if (COMM_RADIUS > 0.0) {
      Coords alternateGoal;
      for(int i = 1; i < nbrNeurons; ++i) {
	 //1st determine the ring segment point
	 const double tdd = Intersection::point_segment_squared_distance(station->coords, cur->coords, cur->next->coords, s, x);
	 const double td = sqrt(tdd);
	 //2nd check if the point is within the communication radius and set new goal
	 if (td > COMM_RADIUS) { //determine new goal
	    alternateGoal.x = station->coords.x + (x.x - station->coords.x) * (COMM_RADIUS_SMALLER / td);
	    alternateGoal.y = station->coords.y + (x.y - station->coords.y) * (COMM_RADIUS_SMALLER / td);
	 } else { //x is within the communication radius
	    alternateGoal = x;
	 }
	 //3 considering the goal and penalty cost
	 const double d = sqrt(alternateGoal.squared_distance(cur->coords));
	 const double penalty = station->penalty;

	 if (step == 0 or (d - penalty ) <= 0) {
	    const double candidate_cost = d - penalty;
	    if (candidate_cost < bestCost) {
	       if (s == 's') { //new point inside the segment
		  newPoint = x; preWinner = cur; winner = 0; //use new point
	       } else {
		  preWinner = 0; winner = s == 'a' ?  cur : cur->next; //use winner
	       }
	       error = d;
	       bestAlternateGoal = alternateGoal;
	       bestCost = candidate_cost;
	       hasWinner = true;
	    }
	 } //end consider it as possible winner candidate
	 cur = cur->next;
      } //end all neurons
   } else {
      for(int i = 1; i < nbrNeurons; ++i) {
	 const double dd = Intersection::point_segment_squared_distance(station->coords, cur->coords, cur->next->coords, s, x);
	 const double d = sqrt(dd);
	 const double penalty = station->penalty;
	 if (step == 0 or (d - penalty ) <= 0) {
	    const double candidate_cost = d - penalty;
	    if (candidate_cost < bestCost) {
	       if (s == 's') { //new point inside the segment
		  newPoint = x; preWinner = cur; winner = 0; //use new point
	       } else {
		  preWinner = 0; winner = s == 'a' ?  cur : cur->next; //use winner
	       }
	       error = d;
	       bestCost = candidate_cost;
	       hasWinner = true;
	    } //end new winner
	 } //end consider it as possible winner candidate
	 cur = cur->next;
      } //end all neurons
   } //end regular penaltyadd winner selection
   if (hasWinner) { //winner has been selected
      if (winner) {
	 if (winner->isInhibited(step) ) { //create new winner
	    preWinner = winner;
	    winner = new SNeuron(preWinner->coords);
	    insertNeuron(preWinner, winner);
	 }
      } else if (preWinner) { //use new point
	 winner = new SNeuron(newPoint);
	 insertNeuron(preWinner, winner);
      }
      errorToGoal = error;
   }
   if (winner) {
      if (COMM_RADIUS > 0.0) {
	 winner->alternateGoal = bestAlternateGoal;
      } else {
	 winner->alternateGoal = station->coords;
      }
      winner->stationOnTour = station->label;
      winner->stationOnTourStep = step;
   }
   return winner;
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::adapt(SNeuron* neuron, const int step) {
   SNeuron* neuronBackward= neuron->prev;
   SNeuron* neuronForward = neuron->next;
   double dd = 1.0;
   const int d = nbrNeurons / NEIGHBORHOOD_FACTOR;
   const int TO = d < (nbrNeurons / 2) ? d : nbrNeurons / 2;
   const Coords& target = neuron->alternateGoal;
   if (neuron->coords.squared_distance(target) > 0) {
      neuron->adapt(target, mi);
      for (int i = 0; i < TO; i++) {
	 const double b = mi*exp((-dd*dd)/(G*G));
	 if (neuronBackward != neuron) {
	    neuronBackward->adapt(target, b);
	    neuronBackward = neuronBackward->prev;
	 }
	 if (neuronForward != neuron) {
	    neuronForward->adapt(target, b);
	    neuronForward = neuronForward->next;
	 }
	 dd += 1.0;
      } //end neighborhood
   }
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::drawRing(int step) {
   if (canvas and neurons) {
      *canvas << canvas::CLEAR << "ring" << "ring";
      *canvas << canvas::LINESTRING << shapeRing;
      double length;
      if (drawRingEnabled) {
	 SNeuron* neuron = neurons;
	 for (int i = 0; i < nbrNeurons; i++) {
	    *canvas << neuron->coords.x << neuron->coords.y;
	    neuron = neuron->next;
	 } //end all neurons
	 *canvas << neurons->coords << canvas::END;
      }
      if (drawNeurons) {
	 *canvas << canvas::POINT << shapeNeuron;
	 SNeuron* neuron = neurons;
	 for(int i = 0; i < nbrNeurons; i++) {
	    if (drawWinners and neuron->stationOnTourStep == step) {
	       *canvas << shapeWinner << neuron->coords << shapeNeuron;
	    } else {
	       *canvas << neuron->coords;
	    }
	    neuron = neuron->next;
	 } //end all neurons
      } else if (drawWinners) {
	 *canvas << canvas::POINT << shapeWinner;
	 SNeuron* neuron = neurons;
	 for(int i = 0; i < nbrNeurons; i++) {
	    if (drawWinners and neuron->stationOnTourStep == step) {
	       *canvas << shapeWinner << neuron->coords;
	    } //end winner
	    neuron = neuron->next;
	 } //end all neurons
      }

      if (drawAlternateGoals and COMM_RADIUS > 0.0) {
	 SNeuron* neuron = neurons;
	 for(int i = 0; i < nbrNeurons; ++i) {
	    if (neuron->stationOnTourStep == step) {
	       *canvas << shapeAlternateGoal << canvas::POINT << neuron->alternateGoal;
	       if (drawNeurons) {  
		  *canvas << Shape::GREEN_LINE << canvas::LINE << neuron->coords << neuron->alternateGoal;
	       }
	    } //end winner
	    neuron = neuron->next;
	 } //end all neurons
	 *canvas << canvas::REDRAW;
      } //end drawing determined goals
      if (drawTourRepresentedByRing) {
	 CoordsVector path;
	 get_ring_path(step, path);
	 if (path.size() > 1) {
	    *canvas 
	       << shapeTourRepresentedByRing << canvas::LINESTRING
	       << path << path.front() << canvas::END;
	    ;
	 }
      }

      if (drawWinnerCommRadius) {
	 *canvas << shapeCommRadius << canvas::ARC;
	 SNeuron* cur = neurons;
	 for(int i = 0; i < nbrNeurons; ++i) {
	    const Coords& ct = cur->coords;
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	    cur = cur->next;
	 } //end all neurons
      }

      if (drawCommRadius) {
	 CoordsVector path;
	 get_ring_path(step, path);
	 *canvas << shapeCommRadius << canvas::ARC;
	 foreach(const Coords& ct, path) {
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	 }
      }
      if (drawCoverageEnable) {
	 CoordsVector path;
	 std::vector<bool> covered;
	 get_ring_path(step, path);
	 get_path_covered_stations(path, covered);
	 *canvas << canvas::ARC << shapeCommRadiusCoverage;
	 foreach(const Coords& ct, path) {
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	 }
	 *canvas 
	    << "stations" << CShape(config.get<std::string>("draw-shape-stations")) 
	    << canvas::POINT;
	 for(int gi = 0; gi < covered.size(); ++gi) {
	    if(covered[gi] == true) {
	       *canvas << FillColor("red");
	    } else {
	       *canvas << FillColor("blue");
	    }
	    *canvas << stations[gi]->coords;
	 }
      } //end drawCoverage
      if (drawCoveredEnable) {
	 CoordsVector path;
	 std::vector<bool> covered;
	 get_ring_path(step, path);
	 get_path_covered_stations(path, covered);
	 CShape shape(config.get<std::string>("draw-shape-stations")); 
	 shape.setFill(false);
	 *canvas << "stations_outline" << canvas::POINT << shape;
	 for(int gi = 0; gi < covered.size(); ++gi) {
	    if(covered[gi] == true) {
	       *canvas << stations[gi]->coords;
	    }
	 } //end all stations
      }
   }
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::drawPath(void) {
   if (canvas) {
      *canvas 
	 << canvas::CLEAR << "path" << "path"
	 << CShape(config.get<std::string>("draw-shape-path")) 
	 << canvas::LINESTRING;
      *canvas << finalPath;
      if (!finalPath.empty()) { *canvas << finalPath.front(); }
      *canvas << canvas::END; 

      if (config.get<bool>("draw-path-vertices")) {
	 *canvas << canvas::POINT << Shape::MAP_VERTEX << finalPath;
      } //end draw-path-vertices
      if (drawCommRadius) {
	 *canvas << shapeCommRadius << canvas::ARC;
	 foreach(const Coords& ct, finalPath) {
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	 }
      } //end drawCommRadius 
   } //end if canvas
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::savePic(int step, bool detail, const std::string& dir_suffix) {
   static int lastStep = step;
   static int i = 0;
   if (lastStep != step) {
      i = 0;
   }
   if (canvas) {
      canvas->redraw();
      std::string dir;
      std::string file = getOutputIterPath(config.get<std::string>("pic-dir") + dir_suffix, dir);
      assert_io(createDirectory(file), "Can not create file in path '" + file + "'");
      std::stringstream ss;
      ss << file << "/" << "iter-" << std::setw(3) << std::setfill('0') << step;
      ss << "-" << std::setw(4) << std::setfill('0') << i;

      std::string suffixes(config.get<std::string>("pic-ext"));
      if (!suffixes.empty()) {
	 std::string::size_type cur = 0;
	 std::string::size_type next;
	 do {
	    next = suffixes.find(',', cur);
	    const std::string& ext = suffixes.substr(cur, next - cur);
	    if (!ext.empty()) {
	       assert_io(canvas->save(ss.str() + "." + ext), "Can create output canvas file '" + file + "'");
	    }
	    cur = next + 1;
	 } while (next != string::npos);
      } else {
	 ss << "." << config.get<std::string>("pic-ext");
	 assert_io(canvas->save(ss.str()), "Can create output canvas file '" + ss.str() + "'");
      }
   }
   lastStep = step;
   i++;
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::releaseStations(void) {
   foreach(SStation* st, stations) { delete st; }
   stations.clear();
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::ring_regenerate(int step) {
   typedef std::vector<SNeuron*> NeuronPtrVector;
   //assuming paths are closed
   SNeuron* cur = neurons;
   //1st determine the winner
   SNeuron* prev_winner = 0;
   for(int i = 0; i < nbrNeurons; ++i) {
      if (cur->isInhibited(step)) {
	 prev_winner = cur;
	 break;
      }
      cur = cur->next;
   }
   NeuronPtrVector del;
   cur = prev_winner->next;

   NeuronPtrVector newNeurons;
   NeuronPtrVector parentNeurons;
   for(int i = 0; i < nbrNeurons; ++i) {
      if (cur) {
	 if (cur->isInhibited(step)) {
	    SNeuron* c = prev_winner;
	    //determine path length from the prev_winner to cur
	    double d = 0.0;
	    while(c != cur) {
	       d += sqrt(c->coords.squared_distance(c->next->coords));
	       if (not c->isInhibited(step)) {
		  del.push_back(c); //mark the neuron to be deleted
	       }
	       c = c->next;
	    }
	    if (cur->coords.squared_distance(prev_winner->coords) > NEURON_COORDS_IDENTITY2) {
	       //determine the center point on the path from prev_winner to cur
	       double l = d / 2.0;
	       c = prev_winner;
	       while(c != cur) {
		  const double d = sqrt(c->coords.squared_distance(c->next->coords));
		  if (l <= d) {
		     const double k = l / d;
		     //place new neuron after the c 
		     const Coords& pt1 = c->coords;
		     const Coords& pt2 = c->next->coords;
		     newNeurons.push_back(new SNeuron(Coords(pt1.x + (pt2.x - pt1.x) * k, pt1.y + (pt2.y - pt1.y) * k)));
		     parentNeurons.push_back(prev_winner);
		     break;
		  } else {
		     l -= d; 
		     c = c->next;
		  }
	       } //end while loop
	    } else {
	       //the current winner is the same as the previous winner 
	       del.push_back(prev_winner);
	    }
	    prev_winner = cur; //update the prev winner
	 } //end cur is inhibited 
	 cur = cur->next;
      } //end cur
   } //end all neurons
   const int N = nbrNeurons;
   foreach(SNeuron* d, del) { removeNeuron(d); } //delete neurons marked for deletion
   for(int i = 0; i < newNeurons.size(); ++i) {
      insertNeuron(parentNeurons[i], newNeurons[i]);
   }
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::load_stations_coords(const std::string& filename) {
   releaseStations();
   Coords pt;
   std::ifstream in(filename.c_str());
   while(in >> pt.x >> pt.y) { 
      stations.push_back(new SStation(stations.size(), pt));
   }
   in.close();
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::load_stations_penalty(const std::string& filename) {
   releaseStations();
   stations.clear();
   int label;
   Coords pt;
   float penalty;
   std::ifstream in(filename.c_str());
   const double factor = config.get<double>("penalty-factor");
   WARN("Currently label of the station in the file is ignored");
   while(in >> label >> pt.x >> pt.y >> penalty) { 
      ASSERT_ARGUMENT(label == stations.size(), "Labels must be precisly incremental");
      stations.push_back(new SStation(stations.size(), pt, penalty / factor));
   }
   in.close();
}

/// - private method -----------------------------------------------------------
CoordsVector& CPCTSPNSom::get_ring_path(int step, CoordsVector& path) const {
   path.clear();
   SNeuron* cur = neurons;
   for(int i = 0; i < nbrNeurons; ++i) {
      if (cur->isInhibited(step) and cur->stationOnTourStep >= 0) {
	 if (COMM_RADIUS > 0.0) { //use the associated alternate goals
	    path.push_back(cur->alternateGoal);
	 } else { //use the associated stations
	    path.push_back(stations[cur->stationOnTour]->coords);
	 }
      }
      cur = cur->next;
   }
   return path;
}

/// - private method -----------------------------------------------------------
double CPCTSPNSom::get_path_length(const CoordsVector& path, bool closed) const {
   double len = 0.0;
   for(int i = 1; i < path.size(); ++i) {
      len += sqrt(path[i-1].squared_distance(path[i]));
   }
   if (closed and path.size() > 1) {
      len += sqrt(path.back().squared_distance(path.front()));
   }
   return len;
}

/// - private method -----------------------------------------------------------
std::vector<bool>& CPCTSPNSom::get_path_covered_stations(const CoordsVector& path, std::vector<bool>& covered, bool closed) const {
   covered.clear();
   covered.resize(stations.size(), false);
   foreach(const Coords& pt, path) {
      for(int gi = 0; gi < stations.size(); ++gi) {
	 if (pt.squared_distance(stations[gi]->coords) <= COMM_RADIUS2) {
	    covered[gi] = true;
	 }
      }
   }
   return covered;
}

/// - private method -----------------------------------------------------------
double CPCTSPNSom::get_path_penalty(const CoordsVector& path, int* coveredStations) const {
   double penalty = 0.0;
   std::vector<bool> covered(stations.size(), false);
   foreach(const Coords& pt, path) {
      for(int si = 0; si < stations.size(); ++si) {
	 if (pt.squared_distance(stations[si]->coords) <= COMM_RADIUS2) {
	    covered[si] = true;
	 }
      }
   }
   for(int i = 0; i < covered.size(); ++i) {
      if(covered[i] == false) {
	 penalty += stations[i]->penalty;
      }
   }
   if (coveredStations) {
      *coveredStations = 0;
      for(int i = 0; i < covered.size(); ++i) {
	 if (covered[i]) {
	    (*coveredStations)++;
	 }
      }
   }
   return penalty;
}

/// - private method -----------------------------------------------------------
void CPCTSPNSom::processPath(void) {
   std::string f = config.get<std::string>("input-path");
   CoordsVector path = imr::CNodeLoader<Coords, CoordsVector>::load(f);

   DEBUG("processPath with nodes: " << path.size());
   Coords bestStation;
   double singleStationCost = -1;

   DEBUG("processPath, the final path has: " << path.size() << " nodes");

   int coveredStations;
   double len = get_path_length(path);
   double penaltyWithRadius = get_path_penalty(path, &coveredStations);
   double penaltyWithoutRadius = get_path_penalty(CoordsVector());

   if (canvas) { //draw
      *canvas 
	 << canvas::CLEAR << "path" << "path"
	 << CShape(config.get<std::string>("draw-shape-path")) 
	 << canvas::LINESTRING
	 << path;
      if (path.size() > 2) { *canvas << path.front(); }
      *canvas << canvas::END;
      if (config.get<bool>("draw-path-vertices")) {
	 *canvas << canvas::POINT << shapeAlternateGoal << path;
      } //end draw-path-vertices
      if (drawCommRadius) {
	 *canvas << shapeCommRadius << canvas::ARC;
	 foreach(const Coords& ct, path) {
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	 }
      } //end drawCommRadius
      if (drawCoverageEnable) {
	 std::vector<bool> covered;
	 get_path_covered_stations(path, covered);
	 *canvas << shapeCommRadiusCoverage << canvas::ARC;
	 foreach(const Coords& ct, path) {
	    *canvas << ct.x << ct.y << COMM_RADIUS << 0.0 << 2*M_PI;
	 }
	 *canvas 
	    << "stations" << CShape(config.get<std::string>("draw-shape-stations")) 
	    << canvas::POINT;
	 for(int si = 0; si < covered.size(); ++si) {
	    if(covered[si] == true) {
	       *canvas << FillColor("red");
	    } else {
	       *canvas << FillColor("gray");
	    }
	    *canvas << stations[si]->coords;
	 }
      } //end drawCoverage
   } //end if canvas

   std::string dir;
   updateResultRecordTimes(); //update timers as load and initilization is outside class
   DEBUG("LOAD_TIME_CPU: " << tLoad.cpuTime());
   DEBUG("INIT_TIME_CPU: " << tInit.cpuTime());
   DEBUG("SAVE_TIME_CPU: " << tSave.cpuTime());
   DEBUG("SOLVE_TIME_CPU: " << tSolve.cpuTime());
   if (SAVE_SETTINGS) {
      saveSettings(getOutputIterPath(config.get<std::string>("settings"), dir));
   }
   if (SAVE_INFO) {
      saveInfo(getOutputIterPath(config.get<std::string>("info"), dir));
   }
   if (SAVE_RESULTS) {
      std::string file = getOutputIterPath(config.get<std::string>("result-path"), dir);
      assert_io(createDirectory(dir), "Can not create file in path'" + file + "'");
      std::ofstream ofs(file.c_str());
      assert_io(not ofs.bad(), "Cannot create path '" + file + "'");
      ofs << std::setprecision(14);
      foreach(const Coords& pt, path) {
	 ofs << pt.x << " " << pt.y << std::endl;
      }
      assert_io(not ofs.bad(), "Error occur during path saving");
      ofs.close();
   }
   saveCanvas();
   CResultLog resLog; //save log
   resLog 
      << result::newcol << "NAME"
      << result::newcol << "METHOD"
      << result::newcol << "RADIUS"
      << result::newcol << "ITER"
      << result::newcol << "LENGTH"
      << CResultLog::AddColumn("PENALTY", "x")
      << CResultLog::AddColumn("COST", "x")
      << CResultLog::AddColumn("PENALTY_NO_RADIUS", "x")
      << CResultLog::AddColumn("COST_NO_RADIUS", "x")
      << CResultLog::AddColumn("COVERED_STATIONS", "x")
      << result::newrec 
      << name << getMethod() << COMM_RADIUS << iter 
      << len << penaltyWithRadius << (len+penaltyWithRadius) << penaltyWithoutRadius << (len+penaltyWithoutRadius)
      << coveredStations 
      << result::endrec;
   std::string log = getOutputPath(config.get<std::string>("results") , dir);
   assert_io(createDirectory(dir), "Can not create directory '" + dir + "'");
   std::fstream out(log.c_str(), std::ios::app | std::ios::out);
   CTextLog t(resLog);
   t.printRecord(resLog.last(), out);
   if (verboseLog) {
      std::stringstream str;
      t.printRecord(resLog.last(), str, false);
      INFO("result log: " << str.str());
   }
   out.close();
   if (!resLog) {
      WARN("Error in resLog:" << resLog.getError());
      resLog.clear();
   }
}

/* end of pctspn_som.cc */
