/*
 * File name: tpctspn.cc
 * Date:      2013/10/13 09:01
 * Author:    Jan Faigl
 */

#include <fstream>
#include <limits.h>

#include <boost/foreach.hpp>

#include <boost/program_options.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include "logging.h"
#include "imr_exceptions.h"
#include "timerN.h"
#include "perf_timer.h"
#include "shapes.h"
#include "guifactory.h"
#include "win_adjust_size.h"

#include "canvasview_coords.h"

#include "pctspn_som.h"
#include "imr_config.h"
#include "boost_args_config.h"

#define foreach BOOST_FOREACH

using imr::logger;

namespace po = boost::program_options;
namespace fs = boost::filesystem;

const std::string PCTSP_SOM_VERSION = "0.1";

using namespace imr;
using namespace imr::gui;

typedef imr::gui::CCanvasBase Canvas;
typedef som::CPCTSPNSom Solver;

typedef std::vector<Coords> CoordsVector;
typedef std::vector<CoordsVector> CoordsVectorVector;

/// ----------------------------------------------------------------------------
/// Program options variables
/// ----------------------------------------------------------------------------
std::string guiType = "none";
std::string problemFile;
std::string problemType;

imr::CConfig guiConfig;
imr::CConfig tspConfig;
std::string canvasOutput = "";

/// ----------------------------------------------------------------------------
/// Global variable
/// ----------------------------------------------------------------------------
imr::gui::CGui * g = 0;
#define GUI(x)  if(gui) { x;}

/// ----------------------------------------------------------------------------
bool parseArgs(int argc, char* argv[]) {
   bool ret = true;
   std::string configFile;
   std::string guiConfigFile;
   std::string loggerCfg = ""; 

   po::options_description desc("General options"); 
   desc.add_options()
      ("help,h", "produce help message")
      ("config,c", po::value<std::string>(&configFile)->default_value(std::string(argv[0]) + ".cfg"), "configuration file")
      ("logger-config,l", po::value<std::string>(&loggerCfg)->default_value(loggerCfg), "logger configuration file")
      ("config-gui", po::value<std::string>(&guiConfigFile)->default_value(std::string(argv[0]) + "-gui.cfg"), "dedicated gui configuration file")
      ("problem", po::value<std::string>(&problemFile), "problem file")
      ("problem-type", po::value<std::string>(&problemType)->default_value("euclid"), "Problem type euclid (TSP) or penalty PCTSP");
   ;
   try {
      po::options_description guiOptions("Gui options");
      imr::gui::CGuiFactory::getConfig(guiConfig);
      imr::gui::CWinAdjustSize::getConfig(guiConfig);
      guiConfig.add<double>("gui-add-x", "add the given value to the loaded goals x coord to determine the canvas size and transformation", 0);
      guiConfig.add<double>("gui-add-y", "add the given value to the loaded goals y coord to determine the canvas size and transformation", 0);
      boost_args_add_options(guiConfig, "", guiOptions);
      guiOptions.add_options() 
	 ("canvas-output", po::value<std::string>(&canvasOutput), "result canvas outputfile")
	 ;

      po::options_description tspOptions("PC-TSP solver options");
      boost_args_add_options(Solver::getConfig(tspConfig), "", tspOptions);

      po::options_description cmdline_options; 
      cmdline_options.add(desc).add(guiOptions).add(tspOptions);

      po::variables_map vm;
      po::store(po::parse_command_line(argc, argv, cmdline_options), vm);
      po::notify(vm);

      std::ifstream ifs(configFile.c_str());
      store(parse_config_file(ifs, cmdline_options), vm);
      po::notify(vm);
      ifs.close();
      ifs.open(guiConfigFile.c_str());
      store(parse_config_file(ifs, cmdline_options), vm);
      po::notify(vm);
      ifs.close();

      if (vm.count("help")) {
	 std::cerr << std::endl;
	 std::cerr << "PC-TSP SOM solver for Euclidean problems ver. " << PCTSP_SOM_VERSION << std::endl;
	 std::cerr << cmdline_options << std::endl;
	 ret = false;
      } 
      if (
	    ret && 
	    loggerCfg != "" && 
	    fs::exists(fs::path(loggerCfg))
	 )  {
	 imr::initLogger("tsp", loggerCfg.c_str());
      } else {
	 imr::initLogger("tsp");
      }
      if (!fs::exists(fs::path(problemFile))) {
	 ERROR("Problem file '" + problemFile + "' does not exists");
	 ret = false;
      }
      if (problemType != "euclid" and problemType != "penalty") {
	 ERROR("Problem type must euclid or penalty!");
	 ret = false;
      }
   } catch (std::exception & e) {
      std::cerr << std::endl;
      std::cerr << "Error in parsing arguments: " << e.what() << std::endl;
      ret = false;
   }
   return ret;
}

/// - private method -----------------------------------------------------------
CoordsVector& load_goals_coords(const std::string& filename, CoordsVector& pts) {
   Coords pt;
   std::ifstream in(problemFile.c_str());
   while(in >> pt.x >> pt.y) { pts.push_back(pt); }
   return pts;
}

/// ----------------------------------------------------------------------------
CoordsVector& load_goals_penalty(const std::string& filename, CoordsVector& pts) {
   int i;
   double cost;
   Coords pt;
   std::ifstream in(problemFile.c_str());
   while(in >> i >> pt.x >> pt.y >> cost) { pts.push_back(pt); }
   return pts;
}

/// ----------------------------------------------------------------------------
/// Main PROGRAM
/// ----------------------------------------------------------------------------
int main(int argc, char* argv[]) {
   Canvas * canvas = 0;
   int ret = -1;
   if (parseArgs(argc, argv)) {
      INFO("Start Logging");
      try {
	 imr::CPerfTimer t("Load problem time real:");
	 CoordsVector pts;
	 imr::CTimerN tLoad;
	 tLoad.start();
	 bool penalty = problemType == "penalty" ? true : false;
	 if (penalty) {
	    load_goals_penalty(problemFile, pts);
	 } else {
	    load_goals_coords(problemFile, pts);
	 }
	 if (tspConfig.get<bool>("draw-communication-radius")) {
	    const double radius = tspConfig.get<double>("communication-radius");
	    const double dx = guiConfig.get<double>("gui-add-x");
	    const double dy = guiConfig.get<double>("gui-add-y");
	    if (radius > 0) {
	       CoordsVector pts1;
	       const int n = pts.size();
	       for(int i = 0; i < n; ++i) {
		  const Coords& pt = pts[i];
		  pts1.push_back(Coords(pt.x + radius + dx, pt.y + radius + dy));
		  pts1.push_back(Coords(pt.x - radius + dx, pt.y - radius - dy));
	       }
	       pts = pts1;
	    }
	 }
	 imr::gui::CWinAdjustSize::adjust(pts, guiConfig);
	 if ((g = gui::CGuiFactory::createGui(guiConfig)) != 0) {
	    INFO("Start gui " +  guiConfig.get<std::string>("gui"));
	    canvas = new Canvas(*g);
	 }
	 tLoad.stop();
	 imr::CTimerN tInit;

	 Solver tsp(tspConfig, problemFile, penalty);
	 tsp.setTimers(tLoad, tInit);
	 tsp.setCanvas(canvas);
	 t.start("TSP solve time: ");
	 tsp.solve();
	 t.stop();
	 INFO("End Logging");
	 if (canvas) {
	    if (canvasOutput.size()) {
	       canvas->save(canvasOutput);
	    }
	    if (!guiConfig.get<bool>("nowait")) {
	       INFO("click to exit");
	       canvas->click();
	    }
	    delete canvas;
	    delete g;
	 }
      } catch (imr::exception& e) {
	 ERROR("Exception " << e.what() << "!");
      } catch (std::exception& e) {
	 ERROR("Runtime error " << e.what() << "!");
      }
      ret = EXIT_SUCCESS;
   }
   imr::shutdownLogger();
   return ret;
}

/* end of tpctspn.cc */
